package edu.kit.ipd.descartes.adaptation;

import java.io.FileInputStream;
import java.io.IOException;
import java.util.Properties;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.xmi.DanglingHREFException;

import edu.kit.ipd.descartes.adaptation.evaluation.WeightingFunctionEvaluator;
import edu.kit.ipd.descartes.adaptation.model.analysis.IModelAnalyzer;
import edu.kit.ipd.descartes.adaptation.model.analysis.pcm.PcmModelAnalyzer;
import edu.kit.ipd.descartes.adaptation.model.dmm.AdaptationProcessModelLoader;
import edu.kit.ipd.descartes.adaptation.model.dmm.DmmModelActionHandler;
import edu.kit.ipd.descartes.adaptation.model.perfdatarepo.IPerformanceDataRepositoryHandler;
import edu.kit.ipd.descartes.adaptation.model.perfdatarepo.PerformanceDataRepositoryHandlerFactory;
import edu.kit.ipd.descartes.mm.adaptation.AdaptationFactory;
import edu.kit.ipd.descartes.mm.adaptation.AdaptationProcess;
import edu.kit.ipd.descartes.mm.adaptation.Event;
import edu.kit.ipd.descartes.mm.adaptation.Strategy;
import edu.kit.ipd.descartes.mm.adaptation.WeightedTactic;
import edu.kit.ipd.descartes.perfdatarepo.Impact;

public class AdaptationControl {

    public enum EventType {
        SLA_VIOLATION("SlaViolatedEvent", "12345"), SCHEDULED_OPTIMIZATION("ScheduledOptimizationEvent",
                "_H_MTgIAnEeKW2vVcg5ekRw"), BOTTLENECK_FOUND("BottleneckFoundEvent", "_VvZEwKEoEeKDl52Xojo4CQ"),
                CUSTOMER_WORKLOAD_UPDATE("CustomerWorkloadUpdate","_XvZEwKEoEeKDl52Xojo4KQ");

        private final String name;
        private final String id;

        EventType(String name, String id) {
            this.name = name;
            this.id = id;
        }

        public static EventType getEventType(String eventTypeName) {
            if (eventTypeName.equals(SLA_VIOLATION.name))
                return SLA_VIOLATION;
            if (eventTypeName.equals(SCHEDULED_OPTIMIZATION.name))
                return SCHEDULED_OPTIMIZATION;
            if (eventTypeName.equals(BOTTLENECK_FOUND.name))
                return BOTTLENECK_FOUND;
            if (eventTypeName.equals(CUSTOMER_WORKLOAD_UPDATE.name))
                return CUSTOMER_WORKLOAD_UPDATE;
            return null;
        }
    }

    private static final String DEFAULT_PROP_FILE_PATH = "./default.properties";
    private static final String PROP_ADAPTATION_PROCESS_XMI_FILE_PATH_NAME = "adaptationprocess.instance";
    private static final String PROP_PERF_DATA_REPO_XMI_FILE_PATH_NAME = "perfdatarepo.instance";
    private static final String PROP_PERF_DATA_REPO_METAMODEL_TYPE = "perfdatarepo.metamodel";
    private static final String PROP_ADAPTATION_PROCESS_NAME = "process.name";
    private static final String PROP_MAX_ITERATIONS = "process.maxiterations";
    private static final String PROP_TRIGGERING_EVENT_TYPE = "event.type";

    private static Logger logger = Logger.getLogger(AdaptationControl.class);

    private String adaptationProcessXmiFilePath = null;
    private String performanceDataRepoXmiFilePath = null;
    private String pdrMetamodelType = null;
    private String eventType = null;

    private AdaptationProcessModelLoader adaptationProcessModelLoader = AdaptationProcessModelLoader.getInstance();
    private IPerformanceDataRepositoryHandler handler;
    private AdaptationProcess adaptationProcess = null;
    private TacticExecutor executor = null;
    private IModelAnalyzer modelAnalyzer = null;
    private WeightingFunctionEvaluator evaluator = null;

    private boolean problemSolved = false;
    private int iteration = 0;
    private int maxIterations = 0;

    /**
     * @param args
     * @throws IOException
     */
    public static void main(String[] args) throws IOException {
        BasicConfigurator.configure();
        Logger.getRootLogger().setLevel(Level.DEBUG);

        AdaptationControl adaptationController = new AdaptationControl();
        // Initialize controller (load models, create Action handler, etc.)
        adaptationController.init();
        Event triggeringEvent = adaptationController.createTriggeringEvent();

        // trigger adaptation process
        adaptationController.doAdaptation(triggeringEvent);
    }

    public void doAdaptation(Event triggeringEvent) {

        // Determine the suitable strategy via the objective
        // which is currently implemented as the first element
        Strategy currentStrategy = AdaptationControl.findTriggeredStrategy(adaptationProcess, triggeringEvent);
        if (currentStrategy == null) {
            logger.error("No strategy found to handle this event!");
            abort();
            return;
        }

        while (!isProblemSolved() && iteration < maxIterations) {
            // Execute the strategie's tactic with the currently highest weight
            WeightedTactic currentTactic = currentStrategy.getTacticWithHighestWeight();

            applyTactic(currentTactic);
            analyzeModel();
            processResults(currentTactic);
            evaluator.evaluate(currentTactic);

            // save results
            try {
                logger.debug("Saving tactics evaluation results...");
                adaptationProcessModelLoader.saveAll();
            } catch (IOException e) {
                if (e.getCause() instanceof DanglingHREFException)
                    logger.error("Error while persisting results because of dangling references. Probably not all changes have been saved");
                else
                    logger.error("Error while persisting evaluation results.", e);
            } 

            iteration++;
        }

        printLogInformation();

    }

    public void init() {
        try {
            // load properties file
            loadProperties(DEFAULT_PROP_FILE_PATH);

            // load required models
            adaptationProcess = adaptationProcessModelLoader.load(URI.createFileURI(adaptationProcessXmiFilePath));
            handler = PerformanceDataRepositoryHandlerFactory.createHandler(pdrMetamodelType);
            handler.load(performanceDataRepoXmiFilePath);

            // set handlers
            executor = new TacticExecutor(new DmmModelActionHandler());
            modelAnalyzer = new PcmModelAnalyzer();
            evaluator = new WeightingFunctionEvaluator();
        } catch (IOException e) {
            logger.error("Error while initializinig controller.", e);
            abort();
        }
    }

    /**
     * Graceful stop the adaptation process.
     */
    public void stop() {
        logger.info("Stopping adaptation process");
        return;
    }

    /**
     * Aborts the adaptation process because of errors
     */
    public void abort() {
        logger.error("Stopping adaptation process due to errors.");
        stop();
    }

    private void loadProperties(String propertiesFilePath) throws IOException {
        Properties properties = new Properties();
        FileInputStream propFile = new FileInputStream(propertiesFilePath);
        properties.load(propFile);
        logger.debug("Loaded properties of " + properties.getProperty(PROP_ADAPTATION_PROCESS_NAME));
        pdrMetamodelType = properties.getProperty(PROP_PERF_DATA_REPO_METAMODEL_TYPE);
        adaptationProcessXmiFilePath = properties.getProperty(PROP_ADAPTATION_PROCESS_XMI_FILE_PATH_NAME);
        performanceDataRepoXmiFilePath = properties.getProperty(PROP_PERF_DATA_REPO_XMI_FILE_PATH_NAME);
        maxIterations = Integer.parseInt(properties.getProperty(PROP_MAX_ITERATIONS));
        eventType = properties.getProperty(PROP_TRIGGERING_EVENT_TYPE).trim();
        logger.debug("Maximum iterations till abort: " + maxIterations);
    }

    // Creates an "artifical event". Will be replaced later when integrated in the system
    private Event createTriggeringEvent() {
        EventType type = EventType.getEventType(eventType);

        if (type == null) {
            logger.error("Could not create event " + eventType);
            abort();
        }

        Event triggeringEvent = AdaptationFactory.eINSTANCE.createEvent();
        triggeringEvent.setName(type.name);
        triggeringEvent.setId(type.id);

        return triggeringEvent;
    }

    private void processResults(WeightedTactic tactic) {

        Impact latestImpact = handler.getImpactAt(iteration);
        // Connect parsed results to the executed tactic
        if (latestImpact != null)
            tactic.setLastImpact(latestImpact);
    }

    private void analyzeModel() {
        modelAnalyzer.analyze();
    }

    private void applyTactic(WeightedTactic weightedTactic) {
        executor.applyTactic(weightedTactic);
    }

    private boolean isProblemSolved() {
        // TODO evaluate the results (problem solved?)
        // TODO Implement correct check of objective
        return problemSolved;
    }

    private void printLogInformation() {
        TacticsHistory hist = TacticsHistory.getInstance();
        hist.printToSysOut();
    }

    private static Strategy findTriggeredStrategy(AdaptationProcess staRepo, Event triggeringEvent) {

        String eventName = triggeringEvent.getName();
        String eventId = triggeringEvent.getId();

        logger.debug("Caught event: " + eventName + " (ID: " + eventId + ")");

        for (Strategy strategy : staRepo.getStrategies()) {
            if (strategy.getTriggeringEvents().getId().equals(eventId)
                    && strategy.getTriggeringEvents().getName().equals(eventName)) {
                logger.debug("Found suitable strategy " + strategy.getName() + " for event "
                        + triggeringEvent.getName());
                return strategy;
            }
        }

        return null;
    }
}
